CREATE_VENV_INFO_SQL = '''
CREATE TABLE IF NOT EXISTS venv_info(
    id INTEGER PRIMARY KEY ASC AUTOINCREMENT,
    project_path VARCHAR(300) NOT NULL,
    venv_name VARCHAR(200) NOT NULL,
    created_date DATE NOT NULL,
    requirement_file VARCHAR(300),
    connect_status VARCHAR(20) NOT NULL,
    last_modified DATETIME NOT NULL,
    UNIQUE(project_path, venv_name)
);
'''

CREATE_VENV_LOG_SQL = '''
CREATE TABLE IF NOT EXISTS venv_log(
    id INTEGER PRIMARY KEY ASC AUTOINCREMENT,
    venv_id INTEGER NOT NULL,
    modify_type VARCHAR(30) NOT NULL,
    modify_time DATETIME NOT NULL,
    modified_by VARCHAR(20) NOT NULL,
    FOREIGN KEY (venv_id) REFERENCES venv_info(id) ON DELETE NO ACTION
);
'''

CREATE_VENV_INSERT_TRIGGER_LOG_SQL = '''
CREATE TRIGGER IF NOT EXISTS venv_insert_info_trigger
AFTER INSERT ON venv_info
FOR EACH ROW
BEGIN 
    INSERT INTO venv_log(venv_id, modify_type, modify_time, modified_by)
    VALUES (NEW.id, 'INSERT VENV', DATETIME('now', 'localtime'), 'USER');
END;
'''

CREATE_VENV_UPDATE_STATUS_TRIGGER_LOG_SQL = '''
CREATE TRIGGER IF NOT EXISTS venv_status_update_trigger
AFTER UPDATE OF connect_status ON venv_info
FOR EACH ROW
BEGIN
    INSERT INTO venv_log(venv_id, modify_type, modify_time, modified_by)
    VALUES (NEW.id, 'UPDATE CONNECTION STATUS', DATETIME('now', 'localtime'), 'SYSTEM');
END;
'''



INSERT_INTO_VENV_INFO_SQL = '''
INSERT INTO venv_info(project_path, venv_name, created_date, requirement_file, connect_status, last_modified) 
VALUES (?, ?, ?, ?, ?, ?);
'''

UPDATE_CONNECT_STATUS_VENV_INFO_SQL = '''
UPDATE venv_info
SET connect_status = ?,
    last_modified = ?
WHERE id = ?;
'''

SELECT_ALL_VENV_SQL = '''
SELECT id AS [VENV ID],
       venv_name AS [VENV],
       created_date AS [CREATED DATE],
       connect_status AS [CONNECTION STATUS],
       project_path AS [PROJECT DIR],
       last_modified AS [LAST MODIFIED]
FROM   venv_info;
'''

SELECT_LATEST_CREATE_DATE_VENV_SQL = '''
SELECT id AS [VENV ID],
       venv_name AS [VENV],
       created_date AS [CREATED DATE],
       connect_status AS [CONNECTION STATUS],
       project_path AS [PROJECT DIR],
       last_modified AS [LAST MODIFIED]
FROM   venv_info
ORDER BY created_date DESC;
'''

SELECT_TOP_LATEST_CREATE_DATA_VENV_SQL = '''
SELECT id AS [VENV ID],
       venv_name AS [VENV],
       created_date AS [CREATED DATE],
       connect_status AS [CONNECTION STATUS],
       project_path AS [PROJECT DIR],
       last_modified AS [LAST MODIFIED]
FROM   venv_info
ORDER BY created_date DESC
LIMIT ?;
'''

SELECT_TOP_LATEST_MODIFIED_DATA_VENV_SQL = '''
SELECT id AS [VENV ID],
       venv_name AS [VENV],
       created_date AS [CREATED DATE],
       connect_status AS [CONNECTION STATUS],
       project_path AS [PROJECT DIR],
       last_modified AS [LAST MODIFIED]
FROM   venv_info
ORDER BY last_modified DESC
LIMIT ?;
'''

SELECT_TOP_EARLIEST_CREATE_DATA_VENV_SQL = '''
SELECT id AS [VENV ID],
       venv_name AS [VENV],
       created_date AS [CREATED DATE],
       connect_status AS [CONNECTION STATUS],
       project_path AS [PROJECT DIR],
       last_modified AS [LAST MODIFIED]
FROM   venv_info
ORDER BY created_date
LIMIT ?;
'''

SELECT_TOP_EARLIEST_MODIFIED_DATA_VENV_VIEW = '''
SELECT id AS [VENV ID],
       venv_name AS [VENV],
       created_date AS [CREATED DATE],
       connect_status AS [CONNECTION STATUS],
       project_path AS [PROJECT DIR],
       last_modified AS [LAST MODIFIED]
FROM   venv_info
ORDER BY last_modified
LIMIT ?;
'''

SELECT_SPECIFIC_VENV_SQL = '''
SELECT id AS [VENV ID],
       venv_name AS [VENV],
       created_date AS [CREATED DATE],
       connect_status AS [CONNECTION STATUS],
       project_path AS [PROJECT DIR],
       last_modified AS [LAST MODIFIED]
FROM   venv_info
WHERE  id = ?;
'''

SELECT_PROJECT_DIR_AND_VENV_SQL = '''
SELECT project_path AS [PROJECT DIR],
       venv_name AS [VENV]
FROM   venv_info
WHERE  project_path = ? AND
       venv_name = ?;
'''

SELECT_NEWEST_INSERT_VENV_ID_SQL = '''
SELECT MAX(id)
FROM   venv_info;
'''

UPDATE_REQUIREMENT_FILE_PATH_SQL = '''
UPDATE venv_info
SET    requirement_file = ?
WHERE  id = ?;
'''

SELECT_SPECIFIC_VENV_CONNECTION_INFO_SQL = '''
SELECT id,
       venv_name,
       project_path,
       connect_status,
       requirement_file
FROM   venv_info
WHERE  id = ?;
'''

INSERT_INSTALL_LIBRARY_LOG_SQL = '''
INSERT INTO venv_log(venv_id, modify_type, modify_time, modified_by)
VALUES (?, 'INSTALL LIBRARIES', DATETIME('now', 'localtime'), 'USER');
'''

INSERT_UNINSTALL_LIBRARY_LOG_SQL = '''
INSERT INTO venv_log(venv_id, modify_type, modify_time, modified_by)
VALUES (?, 'UNINSTALL LIBRARIES', DATETIME('now', 'localtime'), 'USER');
'''

SELECT_ALL_LOG_SQL = '''
SELECT id AS [LOG ID],
       venv_id AS [AFFECTED VENV ID],
       modify_type AS [TYPE OF MODIFICATION],
       modify_time AS [TIME OF MODIFICATION],
       modified_by AS [MODIFIED BY]
FROM   venv_log;
'''

SELECT_LATEST_LOG_SQL = '''
SELECT id AS [LOG ID],
       venv_id AS [AFFECTED VENV ID],
       modify_type AS [TYPE OF MODIFICATION],
       modify_time AS [TIME OF MODIFICATION],
       modified_by AS [MODIFIED BY]
FROM   venv_log
ORDER BY id DESC
LIMIT ?;
'''

SELECT_EARLIEST_LOG_SQL = '''
SELECT id AS [LOG ID],
       venv_id AS [AFFECTED VENV ID],
       modify_type AS [TYPE OF MODIFICATION],
       modify_time AS [TIME OF MODIFICATION],
       modified_by AS [MODIFIED BY]
FROM   venv_log
ORDER BY id
LIMIT ?;
'''




from .utils.ui_controler import menu_screen
from .general.ui_helper import introduction_screen, print_current_page
import warnings



def main():
    
    warnings.filterwarnings("ignore")

    end_program = False

    while not end_program:
        introduction_screen()
        print_current_page('menu')
        end_program = menu_screen()








import re

def str_to_list(text):
    pattern = r'\s'
    pattern = re.compile(pattern, re.X)
    no_white_space_str = re.sub(pattern, '', text)
    str_ls = no_white_space_str.split(',')
    return str_ls

def list_to_str(word_ls, sep = ' '):
    result = ''
    for word in word_ls:
        result = result + word + sep
    result.strip()
    return result









from .all_system_control import *
from .system_control_protocol import *
from .string_processing import *


from . import OS
from .ANSI_color import get_color_str
# from .system_check import get_this_project_version, refresh_check, check_dir_connectivity
from .ANSI_line_modification import *
import inquirer
from inquirer.themes import GreenPassion
    


def menu_screen():
    print(f'              {get_color_str('MENU', 'GREEN')}                     ')
    print()
    menu_choice = [inquirer.List('menu_choice', message = 'Options', choices = [
        '1. Create New Project',
        '2. View Created Virtual Enviroment',
        '3. View Activity Log',
        '4. Exit'
    ], default = '1. Create New Project')]
    user_answer = inquirer.prompt(menu_choice, theme = GreenPassion())['menu_choice']



    end_program = menu_choice_process(user_answer[0])


    if user_answer[0] == '4':
        OS.clear_screen()
        return True
    
    return end_program


from ..utils import OS
from ..utils.system_check import refresh_check, get_this_project_version

from .ANSI_color import get_color_str

def introduction_screen():
    refresh_check()
    OS.clear_screen()
    print('_______________________________________')
    print(f'        {get_color_str('Python VENV Generator', 'GREEN')}     ')
    print()
    print(f' This Version:\t{get_this_project_version()}')
    print(' Created By:\tNguyen Tuan')
    print(f' Your OS:\t{OS.get_os_name()}')
    print('_______________________________________')

def print_current_page(current_page_name: str) -> None:
    print(f'Current Page: ' + get_color_str(current_page_name, 'MANGRETA'))
    print()


import os
import subprocess
# from commands.window_command import *
from ..commands.window_command import *


def get_os_name():
    return 'Windows'

def clear_screen():
    return_status = os.system(clear_screen_command)

def create_venv(project_directory, project_name):
    # subprocess.run([
    #     'cd', project_directory, '&&', 
    #     'python', '-m', 'venv', project_name
    #     ], shell = True)
    subprocess.run([
    'python', '-m', 'venv', f'{project_directory}\\{project_name}'
    ], shell = True)


def install_libraries(project_directory, venv, libraries):
    # os.system(f'cd "{project_directory}"\n' + venv + '\\Scripts\\activate.bat\npip install ' + libraries)

    subprocess.run([
        f'{project_directory}\\{venv}\\Scripts\\pip', 'install', libraries
        ], shell = True)
    
def install_libraries_with_requirement(project_directory, venv, requirement_file):
    subprocess.run([
        f'{project_directory}\\{venv}\\Scripts\\pip', 'install', '-r', f'{requirement_file}'
        ], shell = True)


def activate_venv(project_directory, venv):
    return subprocess.run([
        'cd',
        f'{project_directory}\\{venv}\\Scripts', '&&',
        'activate.bat'
    ], capture_output = True, text = True, shell = True).returncode


def install_system_library(library):
    subprocess.run(['pip', 'install', library], shell = True)

def check_venv(project_directory, venv):
    return os.path.exists(f'{project_directory}\\{venv}\\Scripts\\activate.bat')

def get_requirement(project_dir, venv):
    return subprocess.run([
        f'{project_dir}\\{venv}\\Scripts\\pip', 'freeze'
    ], shell = True, capture_output = True, text = True).stdout

def uninstall_libraries(project_directory, venv, libraries):
    subprocess.run([
        f'{project_directory}\\{venv}\\Scripts\\pip', 'uninstall', libraries, '-y'
    ], shell = True)




from ..general.ANSI_color import get_color_str
from ..general.ui_helper import introduction_screen, print_current_page
from ..general.system_control_protocol import GO_BACK_TO_VIEW_VENV, NO_LIBRARIES
from ..general.string_processing import str_to_list, list_to_str

from ..utils.system_check import check_venv_existence_with_id, check_libraries

from ..utils.requirement_manager import check_installed_libraries

import inquirer
from inquirer.themes import GreenPassion
import pandas as pd
from datetime import datetime



OPTION3_NAVIGATION     = 'menu > view logs'
OPTION3_1_NAVIGATION   = f'{OPTION3_NAVIGATION} > view all logs'
OPTION3_2_NAVIGATION   = f'{OPTION3_NAVIGATION} > view latest logs'
OPTION3_3_NAVIGATION   = f'{OPTION3_NAVIGATION} > view earliest logs'

OPTION3_HEADER     = f'           {get_color_str('View Logs', 'GREEN')}'
OPTION3_1_HEADER   = f'         {get_color_str('View All Logs', 'GREEN')}'
OPTION3_2_HEADER   = f'       {get_color_str('View Latest Logs', 'GREEN')}'
OPTION3_3_HEADER   = f'      {get_color_str('View Earliest Logs', 'GREEN')}'


def print_option3_page(navigation: str, header: str):
    introduction_screen()
    print_current_page(navigation)
    print(header)
    print()

def display_log_menu():
    print_option3_page(OPTION3_NAVIGATION, OPTION3_HEADER)

    view_log_choice = [inquirer.List('view_log_choice', message = 'Options', choices = [
            '1. View All Logs',
            '2. View Latest Logs',
            '3. View Earliest Logs',
            '4. Back',
    ], default = '1. View All Logs')]

    user_answer = inquirer.prompt(view_log_choice, theme = GreenPassion())['view_log_choice']
    return user_answer[0]

def input_no_of_row(navigation: str, header: str):
    valid_answer = False
    first_time = True

    while not valid_answer:
        print_option3_page(navigation, header)
        print(get_color_str('- Please input the number of rows you want to be displayed', 'GREEN'))
        print(get_color_str('- Default is 10 (press Enter in your answer)', 'GREEN'))
        print()

        if not first_time:
            print(get_color_str('Please input a valid answer', 'RED'))

        first_time = False

        try:
            no_of_rows = input('Number of Rows to be Displayed (Default 10): ')
            if no_of_rows == '' or no_of_rows == None:
                return 10
            else:
                no_of_rows = int(no_of_rows)
            return no_of_rows
        except ValueError:
            pass

def display_df(navigation: str, header: str, df: pd.DataFrame):
    print_option3_page(navigation, header)
    print(df.to_markdown())
    print()
    print()
    print(f'<-- Press Enter to go back to {get_color_str('view venv filters', 'MANGRETA')} menu.')
    input()


def display_all_log(df):
    display_df(OPTION3_1_NAVIGATION, OPTION3_1_HEADER, df)

def input_number_of_row_latest_log():
    return input_no_of_row(OPTION3_2_NAVIGATION, OPTION3_2_HEADER)

def display_latest_log(df):
    display_df(OPTION3_2_NAVIGATION, OPTION3_2_HEADER, df)

def input_number_of_row_earliest_log():
    return input_no_of_row(OPTION3_3_NAVIGATION, OPTION3_3_HEADER)

def display_earliest_log(df):
    display_df(OPTION3_3_NAVIGATION, OPTION3_3_HEADER, df)


import sqlite3
import pandas as pd
import types
from datetime import datetime

# from utils.sql_command import *

from .package_directory_manager import get_database_path, create_database_dir
from .sql_command import *
from .ANSI_color import *

'''
Database:

venv_info(id, project_path, venv_name, created_date, requirement_file, connect_status)

'''

def check_database() -> None:
    # If the database directory don't exist this function will create one, else pass
    create_database_dir()
    con = sqlite3.connect(get_database_path())
    cur = con.cursor()

    # If the require tables are not created, these functions will create them
    cur.execute(CREATE_VENV_INFO_SQL)
    cur.execute(CREATE_VENV_LOG_SQL)
    cur.execute(CREATE_VENV_INSERT_TRIGGER_LOG_SQL)
    cur.execute(CREATE_VENV_UPDATE_STATUS_TRIGGER_LOG_SQL)
    con.close()

# It is RECOMMENDED to use this function to safely connect to the database
def connect_check_database() -> sqlite3.Connection:
    # This will return the connection to the database but it will check if the database is availabe first

    check_database()
    con = sqlite3.connect(get_database_path())
    return con

def insert_data_into_venv_info(*args, **kwargs):
    con = connect_check_database()
    cur = con.cursor()
    cur.execute(INSERT_INTO_VENV_INFO_SQL, (kwargs['project_path'], kwargs['venv_name'], kwargs['created_date'], kwargs['requirement_file'], kwargs['connect_status'], kwargs['last_modified']))
    con.commit()
    con.close()

def update_data_venv_info(*args, **kwargs):
    con = connect_check_database()
    cur = con.cursor()
    cur.execute(UPDATE_CONNECT_STATUS_VENV_INFO_SQL, (kwargs['connect_status'], kwargs['last_modified'], kwargs['venv_id']))
    con.commit()
    con.close()


def check_venv_connection(system_control: types.ModuleType, *args, **kwargs):
    con = connect_check_database()
    cur = con.cursor() 
    result = cur.execute(SELECT_ALL_VENV_SQL)
    venv_info = result.fetchall()
    for i in range(len(venv_info)):
        venv_id, venv_name, _, connect_status, project_path, _ = venv_info[i]
        # code_status = system_control.activate_venv(project_path, venv_name)
        # if code_status != 0 and connect_status != get_color_str('no', 'RED'):
        #     update_data_venv_info(con, connect_status = get_color_str('no', 'RED'), venv_name = venv_name, last_modified = datetime.now())
        if not system_control.check_venv(project_path, venv_name) and connect_status != get_color_str('no', 'RED'):
            update_data_venv_info(connect_status = get_color_str('no', 'RED'), venv_id = venv_id, last_modified = datetime.now())
        elif system_control.check_venv(project_path, venv_name):
            pass
    con.close()

def select_all_venv():
    con = connect_check_database()
    df = pd.read_sql(SELECT_ALL_VENV_SQL, con)
    con.close()
    return df

def select_top_latest_create_data_venv(no_of_rows: int) -> pd.DataFrame:
    con = connect_check_database()
    no_of_rows = str(no_of_rows)

    df = pd.read_sql(SELECT_TOP_LATEST_CREATE_DATA_VENV_SQL.replace('?', no_of_rows), con)
    con.close()
    return df

def select_top_latest_modified_data_venv(no_of_rows: int) -> pd.DataFrame:
    con = connect_check_database()
    no_of_rows = str(no_of_rows)

    df = pd.read_sql(SELECT_TOP_LATEST_MODIFIED_DATA_VENV_SQL.replace('?', no_of_rows), con)
    con.close()
    return df

def select_top_earliest_create_data_venv(no_of_rows: int) -> pd.DataFrame:
    con = connect_check_database()
    no_of_rows = str(no_of_rows)

    df = pd.read_sql(SELECT_TOP_EARLIEST_CREATE_DATA_VENV_SQL.replace('?', no_of_rows), con)
    con.close()
    return df

def select_top_earliest_modified_data_venv(no_of_rows: int) -> pd.DataFrame:
    con = connect_check_database()
    no_of_rows = str(no_of_rows)

    df = pd.read_sql(SELECT_TOP_EARLIEST_MODIFIED_DATA_VENV_VIEW.replace('?', no_of_rows), con)
    con.close()
    return df

def select_specific_venv(venv_id: int) -> pd.DataFrame:
    con = connect_check_database()
    venv_id = str(venv_id)

    df = pd.read_sql(SELECT_SPECIFIC_VENV_SQL.replace('?', venv_id), con)
    con.close()
    return df

def check_project_dir_and_venv(project_dir, venv):
    con = connect_check_database()
    cur = con.cursor()
    result = cur.execute(SELECT_PROJECT_DIR_AND_VENV_SQL, (project_dir, venv))
    project_venv = result.fetchall()
    con.close()
    if len(project_venv) > 0:
        return True
    else:
        return False
    
def select_the_latest_inserted_venv_id() -> int:
    con = connect_check_database()
    cur = con.cursor()
    result = cur.execute(SELECT_NEWEST_INSERT_VENV_ID_SQL)
    id = result.fetchall()[0][0]
    con.close()
    return id

def update_requirement(venv_id, requirement_path):
    con = connect_check_database()
    cur = con.cursor()
    cur.execute(UPDATE_REQUIREMENT_FILE_PATH_SQL, (requirement_path, venv_id))
    con.commit()
    con.close()

def check_venv_existence_with_id(venv_id):
    con = connect_check_database()
    cur = con.cursor()
    result = cur.execute(SELECT_SPECIFIC_VENV_SQL, (venv_id,))
    if len(result.fetchall()) > 0:
        return True
    else:
        return False
    
def get_specific_venv_connection(venv_id) -> tuple[bool, str, str, str, str]:
    con = connect_check_database()
    cur = con.cursor() 
    result = cur.execute(SELECT_SPECIFIC_VENV_CONNECTION_INFO_SQL, (venv_id,))
    venv_info = result.fetchall()
    _, venv_name, project_path, connect_status, requirement_file = venv_info[0]
    con.close()


    if connect_status == get_color_str('no', 'RED'):
        return False, project_path, venv_name, requirement_file
    else:
        return True, project_path, venv_name, requirement_file
    
def insert_install_edit_venv_log(venv_id):
    con = connect_check_database()
    cur = con.cursor()
    cur.execute(INSERT_INSTALL_LIBRARY_LOG_SQL, (venv_id,))
    con.commit()
    con.close()

def select_all_venv_list() -> list[tuple]:
    con = connect_check_database()
    cur = con.cursor()
    result = cur.execute(SELECT_ALL_VENV_SQL)
    venv_info = result.fetchall()
    return venv_info

def insert_unistall_edit_venv_log(venv_id):
    con = connect_check_database()
    cur = con.cursor()
    cur.execute(INSERT_UNINSTALL_LIBRARY_LOG_SQL, (venv_id,))
    con.commit()
    con.close()

def select_all_log():
    con = connect_check_database()
    df = pd.read_sql(SELECT_ALL_LOG_SQL, con)
    con.close()
    return df

def select_latest_log(no_of_rows):
    con = connect_check_database()
    df = pd.read_sql(SELECT_LATEST_LOG_SQL.replace('?', str(no_of_rows)), con)
    con.close()
    return df

def select_earliest_log(no_of_rows):
    con = connect_check_database()
    df = pd.read_sql(SELECT_EARLIEST_LOG_SQL.replace('?', str(no_of_rows)), con)
    con.close()
    return df


from ..general.ui_helper import introduction_screen, print_current_page
from ..general.sha256_generator import sha256_generator
from ..general.ANSI_color import get_color_str
from ..general.ANSI_line_modification import GO_UP_ONE_LINE
from ..general.system_control_protocol import GO_BACK_TO_MENU_PAGE, NO_REQUIREMENTS_FILE, NO_LIBRARIES
from ..general.string_processing import str_to_list, list_to_str

from ..utils.system_check import check_dir_connectivity, check_project_dir_and_venv, check_libraries

OPTION1_NAVIGATION = 'menu > create new venv'


def print_option1_page():
    introduction_screen()
    print_current_page(OPTION1_NAVIGATION)
    print(f'            {get_color_str('Create New VENV', 'GREEN')}            ')
    print()


def input_project_directory() -> str:

    first_time = True
    project_directory = None

    while first_time or (not first_time and (project_directory == None or project_directory == '')):
        print_option1_page()

        if not first_time:
            print(f'{get_color_str('Please fill in the valid Directory Name', 'RED')}')

        first_time = False
        
        sha_256_str = sha256_generator()
        print(f'sha256: {sha_256_str}')
        print('Paste this to your answer to go back to main menu')
        print()
        print()
        print()
        project_directory = input(f'Libraries / Dependencies in requirements.txt (Y/n):{GO_UP_ONE_LINE}Project Name / Venv Name:{GO_UP_ONE_LINE}Project Directory: ')
        print()
        print()
        print()

        if project_directory == sha_256_str:
            return GO_BACK_TO_MENU_PAGE

        else:
            directory_exist = check_dir_connectivity(project_directory)
            if not directory_exist:
                project_directory = None

    return project_directory

def input_venv_name(project_directory) -> str:

    first_time = True

    venv_name = None

    valid_project_dir_and_venv = True

    while first_time or (not first_time and (venv_name == None or venv_name == '' or not valid_project_dir_and_venv)):
        print_option1_page()
        
        if not first_time and (venv_name == None or venv_name == ''):
            print(f'{get_color_str('Please fill in the Venv Name', 'RED')}')

        if not first_time and not valid_project_dir_and_venv:
            print(f'{get_color_str('You CANNOT create the same VENV in the same Project Directory', 'RED')}')

        first_time = False
    
        sha_256_str = sha256_generator()
        print(f'sha256: {sha_256_str}')
        print('Paste this to your answer to go back to main menu')
        print()
        print(f'Project Directory: {project_directory}')
        print()

        venv_name = input(f'Libraries / Dependencies in requirements.txt (Y/n):{GO_UP_ONE_LINE}Project Name / Venv Name: ')
        print()
        print()

        if venv_name == sha_256_str:
            return GO_BACK_TO_MENU_PAGE
        elif check_project_dir_and_venv(project_directory, venv_name):
            valid_project_dir_and_venv = False
        else:
            return venv_name
        
def input_requirement_file(project_directory, venv_name) -> str:
    first_time = True
    yes_no_requirement = None

    while first_time or (not first_time and (yes_no_requirement == None or yes_no_requirement == '')):
        print_option1_page()

        if not first_time:
            print(f'{get_color_str('Please fill in the Valid Answer', 'RED')}')

        first_time = False


        sha_256_str = sha256_generator()
        print(f'sha256: {sha_256_str}')
        print('Paste this to your answer to go back to main menu')
        print()
        print(f'Project Directory: {project_directory}')
        print(f'Project Name / Venv Name: {venv_name}')
        yes_no_requirement = input(f'Libraries / Dependencies in requirements.txt (Y/n): ')

        if yes_no_requirement == sha_256_str:
            return GO_BACK_TO_MENU_PAGE
        
        if yes_no_requirement not in ('Y', 'y', 'N', 'n'):
            yes_no_requirement = None
        
    if yes_no_requirement == 'N' or yes_no_requirement == 'n':
        return NO_REQUIREMENTS_FILE
        
    if yes_no_requirement == 'Y' or yes_no_requirement == 'y':

        first_time = True
        requirement_file = None

        while first_time or (not first_time and requirement_file == None):
            print_option1_page()

            if not first_time:
                print(f'{get_color_str('Please fill in the Valid Requirement File Path', 'RED')}')

            first_time = False

            sha_256_str = sha256_generator()
            print(f'sha256: {sha_256_str}')
            print('Paste this to your answer to go back to main menu')
            print()
            print(f'Project Directory: {project_directory}')
            print(f'Project Name / Venv Name: {venv_name}')
            print(f'Libraries / Dependencies in requirements.txt (Y/n): {yes_no_requirement}')
            requirement_file = input('requirements.txt path: ')


            if requirement_file == None or requirement_file == '':
                return NO_REQUIREMENTS_FILE
            
            elif requirement_file == sha_256_str:
                return GO_BACK_TO_MENU_PAGE
            
            elif check_dir_connectivity(requirement_file):
                return requirement_file
            
            else:
                requirement_file = None


def input_libraries(project_directory, venv_name, requirement_file):
    if requirement_file != NO_REQUIREMENTS_FILE:
        return None
    
    first_time = True
    yes_no_libraries = None

    while first_time or (not first_time and (yes_no_libraries == None or yes_no_libraries == '')):
        print_option1_page()

        if not first_time:
            print(f'{get_color_str('Please fill in the Valid Answer', 'RED')}')

        first_time = False

        sha_256_str = sha256_generator()
        print(f'sha256: {sha_256_str}')
        print('Paste this to your answer to go back to main menu')
        print()
        print(f'Project Directory: {project_directory}')
        print(f'Project Name / Venv Name: {venv_name}')
        print(f'Libraries / Dependencies in requirements.txt (Y/n): N')
        yes_no_libraries = input('Set Up Venv with Libraries (Y/n): ')

        if yes_no_libraries == sha_256_str:
            return GO_BACK_TO_MENU_PAGE
        
        if yes_no_libraries not in ('Y', 'y', 'N', 'n'):
            yes_no_libraries = None

    if yes_no_libraries == 'N' or yes_no_libraries == 'n':
        return NO_LIBRARIES
        
    if yes_no_libraries == 'Y' or yes_no_libraries == 'y':

        first_time = True
        libraries = None
        valid_libraries = True
        available = None
        not_available = None

        while first_time or (not first_time and (not valid_libraries)):
            print_option1_page()

            if not first_time and not valid_libraries:
                not_available = list_to_str(not_available, sep = ', ')
                print(get_color_str(f'These are not availabe libraries: {not_available}', 'RED'))
                print(get_color_str('Please type in the libraries again: ', 'RED'))
                not_available = None
                available = None


            first_time = False

            sha_256_str = sha256_generator()
            print(f'sha256: {sha_256_str}')
            print('Paste this to your answer to go back to main menu')
            print()
            print(f'Project Directory: {project_directory}')
            print(f'Project Name / Venv Name: {venv_name}')
            print(f'Libraries / Dependencies in requirements.txt (Y/n): N')
            print(f'Set Up Venv with Libraries (Y/n): {yes_no_libraries}')
            libraries = input('Libraries (Separate by ","): ')

            if libraries == None or libraries == '':
                return NO_LIBRARIES
            
            elif libraries == sha_256_str:
                return GO_BACK_TO_MENU_PAGE
            
            libraries = str_to_list(libraries)
            available, not_available = check_libraries(libraries)

            if len(not_available) > 0:
                valid_libraries = False
                libraries = None
            else:
                available = list_to_str(available, ' ')
                return available


            
def input_venv_info() -> tuple[str, str, str, list]:

    project_directory = input_project_directory()
    if project_directory == GO_BACK_TO_MENU_PAGE:
        return GO_BACK_TO_MENU_PAGE, None, None, None

    venv_name = input_venv_name(project_directory)
    if venv_name == GO_BACK_TO_MENU_PAGE:
        return None, GO_BACK_TO_MENU_PAGE, None, None

    requirement_file = input_requirement_file(project_directory, venv_name)
    if requirement_file == GO_BACK_TO_MENU_PAGE:
        return None, None, GO_BACK_TO_MENU_PAGE, None

    libraries = input_libraries(project_directory, venv_name, requirement_file)
    if libraries == GO_BACK_TO_MENU_PAGE:
        return None, None, None, GO_BACK_TO_MENU_PAGE
    

    return project_directory, venv_name, requirement_file, libraries




    







from ..general.system_control_protocol import *
from ..general.ANSI_color import *
from ..optionui_modules import option1ui, option2ui, option3ui

from .database_manager import (
    insert_data_into_venv_info, 
    select_all_venv,
    select_top_latest_create_data_venv,
    select_top_latest_modified_data_venv,
    select_top_earliest_create_data_venv,
    select_top_earliest_modified_data_venv,
    select_specific_venv,
    update_requirement,
    update_data_venv_info,
    insert_install_edit_venv_log,
    insert_unistall_edit_venv_log,
    select_all_log,
    select_latest_log,
    select_earliest_log
)

from .requirement_manager import requirement_generator, update_install_requirement

from .system_check import refresh_check, get_specific_venv_connection

from datetime import date, datetime


import pandas as pd

from . import OS






def option1():

    project_directory, venv_name, requirement_file, libraries = option1ui.input_venv_info()
    if GO_BACK_TO_MENU_PAGE in (project_directory, venv_name, requirement_file, libraries):
        return False

    
    if requirement_file == NO_REQUIREMENTS_FILE:

        OS.create_venv(project_directory, venv_name)
        insert_data_into_venv_info(
            project_path = project_directory, 
            venv_name = venv_name, 
            created_date = date.today(), 
            requirement_file = None, 
            connect_status = get_color_str('yes', 'GREEN'), 
            last_modified = datetime.now()
        )

        if libraries != NO_LIBRARIES:
            OS.install_libraries(project_directory, venv_name, libraries)
        
    
    else:
        OS.create_venv(project_directory, venv_name)
        OS.install_libraries_with_requirement(project_directory, venv_name, requirement_file)
        insert_data_into_venv_info(
            project_path = project_directory, 
            venv_name = venv_name, 
            created_date = date.today(), 
            requirement_file = requirement_file, 
            connect_status = get_color_str('yes', 'GREEN'), 
            last_modified = datetime.now()
        )
        
    # else:
    #     OS.create_venv(project_directory, venv_name)
    #     insert_data_into_venv_info(
    #         project_path = project_directory, 
    #         venv_name = venv_name, 
    #         created_date = date.today(), 
    #         requirement_file = None, 
    #         connect_status = get_color_str('yes', 'GREEN'), 
    #         last_modified = datetime.now()
    #     )


    # TODO: check on this thing
    venv_id, requirement_path = requirement_generator(project_directory, venv_name)
    update_requirement(venv_id, requirement_path)
    return False

def option2_1():
    refresh_check()
    all_venv_df = select_all_venv()
    option2ui.display_all_view(all_venv_df)
    return False
    
def option2_2_1():
    no_of_rows = option2ui.input_no_of_rows_latest_create()
    latest_create_venv_df = select_top_latest_create_data_venv(no_of_rows)
    option2ui.display_latest_create_date(latest_create_venv_df)
    return False

def option2_2_2():
    no_of_rows = option2ui.input_no_of_rows_latest_modified()
    latest_modified_venv_df = select_top_latest_modified_data_venv(no_of_rows)
    option2ui.display_latest_modified_time(latest_modified_venv_df)
    return False

def option2_2_3():
    no_of_rows = option2ui.input_no_of_rows_earliest_create()
    earliest_create_venv_df = select_top_earliest_create_data_venv(no_of_rows)
    option2ui.display_earliest_create_date(earliest_create_venv_df)
    return False

def option2_2_4():
    no_of_rows = option2ui.input_no_rows_earliest_modified()
    earliest_modified_venv_df = select_top_earliest_modified_data_venv(no_of_rows)
    option2ui.display_earliest_modified_time(earliest_modified_venv_df)
    return False

    
def option2_2():
    refresh_check()

    option2_2_lookup_table = {
        '1': option2_2_1,
        '2': option2_2_2,
        '3': option2_2_3,
        '4': option2_2_4,
        '5': (lambda : True)
    }

    to_view_menu = False

    while not to_view_menu:
        user_choice = option2ui.display_with_filter_menu()
        to_view_menu = option2_2_lookup_table[user_choice]()
    return False

def option2_3():
    venv_id = option2ui.input_venv_id_for_view()
    if venv_id == GO_BACK_TO_VIEW_VENV:
        return False
    else:
        venv_detail = select_specific_venv(venv_id)
        option2ui.display_specific_view(venv_detail)


def option2_4():
    option2_4_lookup_table = {
        '1': option2ui.venv_edit_install_library,
        '2': option2ui.venv_edit_uninstall_library
    }


    venv_id = option2ui.input_venv_id_for_edit()

    if venv_id == GO_BACK_TO_VIEW_VENV:
        return False
    else:
        connected, project_dir, venv_name, requirement_file = get_specific_venv_connection(venv_id)

        if not connected and option2ui.recover_venv():
            OS.create_venv(project_dir, venv_name)
            OS.install_libraries_with_requirement(project_dir, venv_name, requirement_file)
            update_data_venv_info(connect_status = get_color_str('yes','GREEN'), last_modified = datetime.now(), venv_id = venv_id)
            

        #TODO: create an option2ui.venv_edit
        # create an option2ui.venv_edit_menu
        else:
            user_choice = option2ui.venv_edit_menu()

            

            libraries = option2_4_lookup_table[user_choice](venv_id)
            
            if libraries == GO_BACK_TO_VIEW_VENV:
                return False
            elif user_choice == '1':
                # print(venv_id, project_dir, venv_name)
                # input()



                OS.install_libraries(project_dir, venv_name, libraries)
                insert_install_edit_venv_log(venv_id)
                update_install_requirement(venv_id)
                return False
            elif user_choice == '2':
                OS.uninstall_libraries(project_dir, venv_name, libraries)
                insert_unistall_edit_venv_log(venv_id)
                update_install_requirement(venv_id)
                return False
            




    
    return False
        

def option2():
    
    option2_lookup_table = {
        '1': option2_1,
        '2': option2_2,
        '3': option2_3,
        '4': option2_4,
        '5': (lambda : True)
    }

    to_main_menu = False

    while not to_main_menu:
        user_choice = option2ui.view_venv_menu()

        to_main_menu = option2_lookup_table[user_choice]()
        
    return False

def option3():

    
    to_main_menu = False

    while not to_main_menu:
        user_choice = option3ui.display_log_menu()
        
        if user_choice == '1':
            all_log = select_all_log()
            to_main_menu = option3ui.display_all_log(all_log)
        elif user_choice == '2':
            no_of_rows = option3ui.input_number_of_row_latest_log()
            latest_log = select_latest_log(no_of_rows)
            to_main_menu = option3ui.display_latest_log(latest_log)
        elif user_choice == '3':
            no_of_rows = option3ui.input_number_of_row_earliest_log()
            earliest_log = select_earliest_log(no_of_rows)
            to_main_menu = option3ui.display_earliest_log(earliest_log)
        elif user_choice == '4':
            to_main_menu = True

    return False



    






from .check_os import check_os

OS = check_os()


from .depedencies_check import check_dependencies
check_dependencies()



print('Importing system check...', end = '                                             \r')
# from . import system_check

# system_check.init_check()

# OS = system_check.check_os()

from .system_check import init_check, check_dir_connectivity, check_project_dir_and_venv, check_libraries

# init_check()


# print('Importing system control...', end = '                                          \r')
# from . import all_system_control

# print('Importing ANSI color...', end = '                                              \r')
# from . import ANSI_color

# print('Importing ANSI line modification...', end = '                                  \r')
# from . import ANSI_line_modification

# print('Importing database manager...', end = '                                        \r')
# from . import database_manager

# print('Importing macos system control...', end = '                                    \r')
# from . import macos_system_control

# print('Importing option manager...', end = '                                          \r')
# from . import option_manager

# print('Importing package directory manager...', end = '                                   \r')
# from . import package_directory_manager

# print('Importing sql command...', end = '                                             \r')
# from . import sql_command

# print('Importing string_processing', end = '                                          \r')
# from . import string_processing

# print('Importing system control protocol...', end = '                                 \r')
# from . import system_control_protocol

print('Importing UI controler...', end = '                                            \r')
from . import ui_controler

# print('Importing Window system control', end = '                                      \r')
# from . import window_system_control




from typing import Literal

ANSI_color_dict = {
    'RED': '91',
    'GREEN': '32',
    'BLUE': '34',
    'YELLOW': '93',
    'MANGRETA': '95',
    'BRIGHT WHITE': '97',
    'WHITE': '37'
}

def get_color_str(text: str, color: Literal['RED', 'GREEN', 'BLUE', 'YELLOW', 'MANGRETA', 'BRIGHT WHITE', 'WHITE']) -> str:
    START = f'\33[{ANSI_color_dict[color]}m'
    END   = '\33[m'
    return START + text + END


def print_color(text: str, color: str) -> None:
    START = f'\33[{ANSI_color_dict[color]}m'
    END   = '\33[m'
    print(START + text + END, end = '')





import platform
from types import ModuleType

def check_os() -> ModuleType:
    if platform.system() == 'Darwin':
        from ..OS_control import macos_system_control as OS
        return OS
    elif platform.system() == 'win32' or platform.system() == 'Windows':
        from ..OS_control import window_system_control as OS
        return OS


import os
# from commands.macos_command import *
from ..commands.macos_command import *

def get_os_name():
    return 'MacOS'

def clear_screen():
    return_status = os.system(clear_screen_command)

def create_venv(project_directory, project_name):
    os.system(f'cd "{project_directory}"\n' + CREATE_VENV + project_name)

def install_libraries(project_directory, venv, libraries):
    os.system(f'cd "{project_directory}"\n' + 'source ' + venv + '/bin/activate\npip install ' + libraries)

def install_libraries_with_requirement(project_directory, venv, requirement_file):
    os.system(f'cd "{project_directory}"\n' + 'source ' + venv + '/bin/activate\npip install -r ' + f'"{requirement_file}"')

def activate_venv(project_directory, venv):
    return os.system(f'cd "{project_directory}"\n' + 'source ' + venv + '/bin/activate\n')

def install_system_library(library):
    os.system(f'pip install {library}')

def check_venv(project_directory, venv):
    return os.path.exists(f'{project_directory}/{venv}/bin/activate')

def get_requirement(project_dir: str, venv: str) -> str:
    return os.popen(f'cd "{project_dir}"\nsource "{venv}/bin/activate"\npip freeze\n').read()

def uninstall_libraries(project_directory, venv, libraries):
    os.system(f'cd "{project_directory}"\nsource {venv}/bin/activate\npip uninstall {libraries} -y')


GO_BACK_TO_MENU_PAGE = 'asgjqpo3irht=0239r1p93q8yrtiogwehjknldfmbnq24[e0dp9touiq3fj1kqrmwefglobwdk.jsnfbvc efgdboqkelt;nrfg,bwopeust;dhjflgknwmepstdofgjblvkwelargjksfhdnw.vearspg]'
NO_REQUIREMENTS_FILE = 'asgdkq3jrhrgqpviuaebfjndgqo8ehiar[ksdgf;v30 [qeiors;tdfghvq3aisdjkthq3p9rtegoui350-r9t8729385yreotufghilsjdnvcm,x q;eoritulgdfhkjnc,mvw4eoirtdgh;flkjsda ]]'
NO_LIBRARIES         = 'asgriojnqenkqe;lkneqrkeqg;lnkefgql;nkds;jew[jebo[keqda;fgibjlvkwesdtgo;fjknblvmseoidtgjf;klbmvwtei[gijafdmvspqregij;fkamdsveowigfknbdlmsrhwbgnsdjlvm /eq]]]'
GO_BACK_TO_VIEW_VENV = 'sjkfd;qygihosajknvcxm,pyruehtqg;jkfadn8030qru[egijof3tuwqhigdsafjkvn9rqheugafdjnbrqtehgafdjln0u etwjiqoadfs;e[ghijofdbslkeuqghfadbjknverhuqgafdskjnv.cxhe]]'


import os

def get_parent_directory(dir):
    return os.path.dirname(dir)

def create_database_dir():
    database_dir = os.path.join(get_parent_directory(get_package_path()), 'database')

    if not os.path.exists(database_dir):
        os.makedirs(database_dir)

def get_package_path():
    package_dir = os.path.dirname(os.path.abspath(__file__))
    return package_dir

def get_database_dir():
    return os.path.join(get_parent_directory(get_package_path()), 'database')

def get_database_path():
    return os.path.join(get_database_dir(), "venv_database.db")

def create_requirement_dir():
    requirement_dir = os.path.join(get_parent_directory(get_package_path()), 'requirements')

    if not os.path.exists(requirement_dir):
        os.makedirs(requirement_dir)

def get_requirement_dir():
    return os.path.join(get_parent_directory(get_package_path()), 'requirements')

def get_requirement_path(venv_id):
    return os.path.joint(get_requirement_dir(), f'{venv_id}.txt')




# from utils.ANSI_color import *
# from utils.system_control_protocol import *
# from utils.option_manager import *
# from utils import macos_system_control as MacOS
# from utils import window_system_control as Windows

# from .system_control_protocol import *
from .option_manager import option1, option2, option3
import random
import hashlib


def sha256_generator():
    return hashlib.sha256(str(random.randint(-100000000000, 10000000000000)).encode('utf-8')).hexdigest()


    

def menu_choice_process(user_answer):
    if user_answer == '1':
        return option1()
    elif user_answer == '2':
        return option2()
    elif user_answer == '3':
        return option3()





from .check_os import check_os
from .ANSI_color import get_color_str, print_color

def check_dependencies():
    OS = check_os()

    try:
        import inquirer
    except ModuleNotFoundError as e:
        print()
        print(f'{get_color_str('inquirer', 'RED')} module have not yet installed')
        print()
        print('Do you want to download it to your machine?')
        print()
        print('The library is Required for the program to run')
        userChoice = input('Your choice (Y/n): ')
        if userChoice == 'Y' or userChoice == 'y':
            OS.install_system_library('inquirer')
            print()
            print('Please rerun the program after installation')
            exit()        
        else:
            print()
            print_color('Cannot process further due to the lack of library', 'RED')
            exit()

    try:
        import pandas
    except ModuleNotFoundError as e:
        print()
        print(f'{get_color_str('pandas', 'RED')} module have not yet installed')
        print()
        print('Do you want to download it to your machine?')
        print()
        print('The library is Required for the program to run')
        userChoice = input('Your choice (Y/n): ')
        if userChoice == 'Y' or userChoice == 'y':
            OS.install_system_library('pandas')
            print()
            print('Please rerun the program after installation')
            exit()        
        else:
            print()
            print_color('Cannot process further due to the lack of library', 'RED')
            exit()

    try:
        import tabulate
    except ModuleNotFoundError as e:
        print()
        print(f'{get_color_str('tabulate', 'RED')} module have not yet installed')
        print()
        print('Do you want to download it to your machine?')
        print()
        print('The library is Required for the program to run')
        userChoice = input('Your choice (Y/n): ')
        if userChoice == 'Y' or userChoice == 'y':
            OS.install_system_library('tabulate')
            print()
            print('Please rerun the program after installation')
            exit()        
        else:
            print()
            print_color('Cannot process further due to the lack of library', 'RED')
            exit()

    try:
        import requests
    except ModuleNotFoundError as e:
        print()
        print(f'{get_color_str('requests', 'RED')} module have not yet installed')
        print()
        print('Do you want to download it to your machine?')
        print()
        print('The library is Required for the program to run')
        userChoice = input('Your choice (Y/n): ')
        if userChoice == 'Y' or userChoice == 'y':
            OS.install_system_library('requests')
            print()
            print('Please rerun the program after installation')
            exit()        
        else:
            print()
            print_color('Cannot process further due to the lack of library', 'RED')
            exit()




from ..general.ANSI_color import get_color_str
from ..general.ui_helper import introduction_screen, print_current_page
from ..general.system_control_protocol import GO_BACK_TO_VIEW_VENV, NO_LIBRARIES
from ..general.string_processing import str_to_list, list_to_str

from ..utils.system_check import check_venv_existence_with_id, check_libraries

from ..utils.requirement_manager import check_installed_libraries

import inquirer
from inquirer.themes import GreenPassion
import pandas as pd
from datetime import datetime



OPTION2_NAVIGATION     = 'menu > view venv details'
OPTION2_1_NAVIGATION   = f'{OPTION2_NAVIGATION} > view all venv'
OPTION2_2_NAVIGATION   = f'{OPTION2_NAVIGATION} > view venv (filtered)'
OPTION2_2_1_NAVIGATION = f'{OPTION2_2_NAVIGATION} > view latest created date'
OPTION2_2_2_NAVIGATION = f'{OPTION2_2_NAVIGATION} > view latest modified time'
OPTION2_2_3_NAVIGATION = f'{OPTION2_2_NAVIGATION} > view earliest created date'
OPTION2_2_4_NAVIGATION = f'{OPTION2_2_NAVIGATION} > view earliest modified time'
OPTION2_3_NAVIGATION   = f'{OPTION2_NAVIGATION} > venv details'
OPTION2_4_NAVIGATION   = f'{OPTION2_NAVIGATION} > venv edit'


OPTION2_HEADER     = f'          {get_color_str('View Venv Details', 'GREEN')}'
OPTION2_1_HEADER   = f'            {get_color_str('View All VENV', 'GREEN')}'
OPTION2_2_HEADER   = f'         {get_color_str('View Venv (filtered)', 'GREEN')}'
OPTION2_2_1_HEADER = f'   {get_color_str('View VENV (Sorted By Latest Created Date)', 'GREEN')}'
OPTION2_2_2_HEADER = f'   {get_color_str('View VENV (Sorted By Latest Modified Time)', 'GREEN')}'
OPTION2_2_3_HEADER = f'  {get_color_str('View VENV (Sorted By Earliest Created Date)', 'GREEN')}'
OPTION2_2_4_HEADER = f'  {get_color_str('View VENV (Sorted By Earliest Modified Time)', 'GREEN')}'
OPTION2_3_HEADER   = f'         {get_color_str('View Specific VENV Details', 'GREEN')}'
OPTION2_4_HEADER   = f'           {get_color_str('Venv Edit', 'GREEN')}'


def print_option2_page(navigation: str, header: str):
    introduction_screen()
    print_current_page(navigation)
    print(header)
    print()

def view_venv_menu():
    print_option2_page(OPTION2_NAVIGATION, OPTION2_HEADER)

    view_choice = [inquirer.List('view_choice', message = 'Options', choices = [
            '1. View All Venv',
            '2. View Venv (with Filter)',
            '3. View A Venv Details',
            '4. Edit Venv',
            '5. Back'
    ], default = '1. View All Venv')]

    user_answer = inquirer.prompt(view_choice, theme = GreenPassion())['view_choice']
    
    return user_answer[0]

def display_all_view(df: pd.DataFrame):
    print_option2_page(OPTION2_1_NAVIGATION, OPTION2_1_HEADER)
    print(df.to_markdown())
    print()
    print()
    print(f'<-- Press Enter to go back to {get_color_str('view venv details', 'MANGRETA')} menu.')
    input()

def display_with_filter_menu():
    print_option2_page(OPTION2_2_NAVIGATION, OPTION2_2_HEADER)

    view_filter_choice = [inquirer.List('view_filter_choice', message = 'Options', choices = [
            '1. View by Latest Created Day',
            '2. View by Latest Modified Time',
            '3. View by Earliest Created Day',
            '4. View by Earliest Modified Time',
            '5. Back'
    ], default = '1. View All Venv')]

    user_answer = inquirer.prompt(view_filter_choice, theme = GreenPassion())['view_filter_choice']
    return user_answer[0]


def input_no_of_row(navigation: str, header: str):
    valid_answer = False
    first_time = True

    while not valid_answer:
        print_option2_page(navigation, header)
        print(get_color_str('- Please input the number of rows you want to be displayed', 'GREEN'))
        print(get_color_str('- Default is 10 (press Enter in your answer)', 'GREEN'))
        print()

        if not first_time:
            print(get_color_str('Please input a valid answer', 'RED'))

        first_time = False

        try:
            no_of_rows = input('Number of Rows to be Displayed (Default 10): ')
            if no_of_rows == '' or no_of_rows == None:
                return 10
            else:
                no_of_rows = int(no_of_rows)
            return no_of_rows
        except ValueError:
            pass

def display_df(navigation: str, header: str, df: pd.DataFrame):
    print_option2_page(navigation, header)
    print(df.to_markdown())
    print()
    print()
    print(f'<-- Press Enter to go back to {get_color_str('view venv filters', 'MANGRETA')} menu.')
    input()


def input_no_of_rows_latest_create() -> int:
    return input_no_of_row(OPTION2_2_1_NAVIGATION, OPTION2_2_1_HEADER)

def display_latest_create_date(df: pd.DataFrame):
    display_df(OPTION2_2_1_NAVIGATION, OPTION2_2_1_HEADER, df)

def input_no_of_rows_latest_modified() -> int:
    return input_no_of_row(OPTION2_2_2_NAVIGATION, OPTION2_2_2_HEADER)

def display_latest_modified_time(df: pd.DataFrame):
    display_df(OPTION2_2_2_NAVIGATION, OPTION2_2_2_HEADER, df)

def input_no_of_rows_earliest_create():
    return input_no_of_row(OPTION2_2_3_NAVIGATION, OPTION2_2_3_HEADER)

def display_earliest_create_date(df: pd.DataFrame):
    display_df(OPTION2_2_3_NAVIGATION, OPTION2_2_3_HEADER, df)

def input_no_rows_earliest_modified() -> int:
    return input_no_of_row(OPTION2_2_4_NAVIGATION, OPTION2_2_4_HEADER)

def display_earliest_modified_time(df: pd.DataFrame):
    display_df(OPTION2_2_4_NAVIGATION, OPTION2_2_4_HEADER, df)

def input_venv_id(navigation: str, header: str) -> int:
    valid_answer = False
    first_time = True
    exist_venv = False

    while not valid_answer or not exist_venv:
        print_option2_page(navigation, header)
        print(get_color_str('- Please input the venv id', 'GREEN'))
        print(get_color_str('- The id must be an integer', 'GREEN'))
        print(get_color_str('- Type "back" to go back to view venv', 'GREEN'))
        print()

        if not first_time and not valid_answer:
            print(get_color_str('Please input a valid answer', 'RED'))

        if not first_time and not exist_venv:
            print(get_color_str('Venv does NOT exist', 'RED'))

        first_time = False

        venv_id = input('VENV id: ')
        if venv_id == 'back':
            return GO_BACK_TO_VIEW_VENV

        try:
            venv_id = int(venv_id)
            valid_answer = True
        except ValueError:
            valid_answer = False
        
        if check_venv_existence_with_id(venv_id):
            return venv_id
        else:
            exist_venv = False
        

def input_venv_id_for_view():
    return input_venv_id(OPTION2_3_NAVIGATION, OPTION2_3_HEADER)

def display_specific_view(df: pd.DataFrame):
    print_option2_page(OPTION2_3_NAVIGATION, OPTION2_3_HEADER)
    print(df.to_markdown())
    print()
    print()
    print(f'<-- Press Enter to go back to {get_color_str('view venv details', 'MANGRETA')} menu.')
    input()

def input_venv_id_for_edit():
    return input_venv_id(OPTION2_4_NAVIGATION, OPTION2_4_HEADER)

def recover_venv():
    print_option2_page(OPTION2_4_NAVIGATION, OPTION2_4_HEADER)
    print(get_color_str('The given venv has lost the connection', 'RED'))
    yes_no_recover = [inquirer.List('yes_no_recover', message = 'Do you want to recover it?', choices = [
            '1. YES',
            '2. NO'
    ], default = '1. View All Venv')]

    user_answer = inquirer.prompt(yes_no_recover, theme = GreenPassion())['yes_no_recover']

    if user_answer[0] == '1':
        return True
    else:
        return False
    
def venv_edit_menu():
    print_option2_page(OPTION2_4_NAVIGATION, OPTION2_4_HEADER)
    venv_edit_option = [inquirer.List('venv_edit_option', message = 'Option:', choices = [
            '1. Install More Libraries',
            '2. Uninstall More Libraries'
    ], default = '1. Install More Libraries')]

    user_answer = inquirer.prompt(venv_edit_option, theme = GreenPassion())['venv_edit_option']
    return user_answer[0]

def venv_edit_install_library(venv_id):
    first_time: bool = True
    no_library: bool = True
    valid_libraries: bool = True
    is_installed: bool = False


    available: list = None
    not_available: list = None

    while first_time or (not first_time and (not valid_libraries or no_library or is_installed)):
        libraries = None
        print_option2_page(OPTION2_4_NAVIGATION, OPTION2_4_HEADER)
        print(get_color_str('- Please type "back" to go back to the view venv details', 'GREEN'))

        if not first_time and not valid_libraries:
            not_available = list_to_str(not_available, sep = ', ')
            print(get_color_str(f'These are not availabe libraries: {not_available}', 'RED'))
            print(get_color_str('Please type in the libraries again: ', 'RED'))
            not_available = None
            available = None

        if not first_time and no_library:
            print(get_color_str('Please insert the libraries name!', 'RED'))

        if not first_time and is_installed:
            print(get_color_str('The libraries have been installed already!', 'RED'))


        first_time = False

        libraries = input('Libraries (Separate by ","): ')


        
        if libraries == 'back':
            return GO_BACK_TO_VIEW_VENV
        
        if libraries == '' or libraries == None:
            no_library = True
            valid_libraries = True
            is_installed = False
        
            
        if libraries != '' and libraries != None:
            no_library = False

            libraries = str_to_list(libraries)
            available, not_available = check_libraries(libraries)

            if len(not_available) > 0:
                valid_libraries = False
                libraries = None
                is_installed = False

            elif len(check_installed_libraries(venv_id, available)[0]) > 0:
                is_installed = True
                no_library = False
                valid_libraries = True
            
            else:
                available = list_to_str(available, ' ')
                return available
            
def venv_edit_uninstall_library(venv_id):
    first_time: bool = True
    no_library: bool = True
    valid_libraries: bool = True
    is_not_installed: bool = False


    available: list = None
    not_available: list = None

    while first_time or (not first_time and (not valid_libraries or no_library or is_not_installed)):
        libraries = None
        print_option2_page(OPTION2_4_NAVIGATION, OPTION2_4_HEADER)
        print(get_color_str('- Please type "back" to go back to the view venv details', 'GREEN'))

        if not first_time and not valid_libraries:
            not_available = list_to_str(not_available, sep = ', ')
            print(get_color_str(f'These are not availabe libraries: {not_available}', 'RED'))
            print(get_color_str('Please type in the libraries again: ', 'RED'))
            not_available = None
            available = None

        if not first_time and no_library:
            print(get_color_str('Please insert the libraries name!', 'RED'))

        if not first_time and is_not_installed:
            print(get_color_str('The libraries have NOT been installed yet!', 'RED'))


        first_time = False

        libraries = input('Libraries (Separate by ","): ')


        
        if libraries == 'back':
            return GO_BACK_TO_VIEW_VENV
        
        if libraries == '' or libraries == None:
            no_library = True
            valid_libraries = True
            is_not_installed = False
        
            
        if libraries != '' and libraries != None:
            no_library = False

            libraries = str_to_list(libraries)
            available, not_available = check_libraries(libraries)

            if len(not_available) > 0:
                valid_libraries = False
                libraries = None
                is_not_installed = False

            elif len(check_installed_libraries(venv_id, available)[1]) > 0:
                is_not_installed = True
                no_library = False
                valid_libraries = True
            
            else:
                available = list_to_str(available, ' ')
                return available















# def input_install_venv_library():
#     print_option2_page(OPTION2_4_NAVIGATION, OPTION2_4_HEADER)



from .ANSI_color import get_color_str, print_color

# All of the imported functions that are not used in this module will be used by the other modules that import this module
from .database_manager import (
    check_database, 
    check_venv_connection, 
    check_project_dir_and_venv, 
    check_venv_existence_with_id, 
    get_specific_venv_connection
)
from .requirement_manager import check_requirement
from ..venvgen_version import __version__
from .depedencies_check import check_os

import os
import requests

# TODO: This function may need to be deleted
def init_check():
    check_database()
    check_requirement()

def refresh_check():
    check_venv_connection(check_os())
    check_requirement()

def get_this_project_version() -> str:
    return __version__

def check_dir_connectivity(dir: str) -> bool:
    return os.path.exists(dir)

def check_libraries(libraries: list) -> tuple[list, list]:
    available = []
    not_available = []

    print('Checking libraries availability...')

    for i in range(len(libraries)):

        status = requests.get(f'https://pypi.org/project/{libraries[i]}/').status_code
        
        if status == 200:
            available.append(libraries[i])

        if status == 404:
            not_available.append(libraries[i])

        percentage = ((i + 1) / len(libraries))
        percentage_100 = percentage * 100

        MAX_BAR_LEN = 50

        percentage_bar_len = int(percentage * MAX_BAR_LEN) - 1

        remaining_bar_len = MAX_BAR_LEN - percentage_bar_len - 1

        percentage_bar = '=' * percentage_bar_len + '>'
        remaining_bar  = '.' * remaining_bar_len


        print(f'Progress: |{percentage_bar}{remaining_bar}| {percentage_100:.2f}%', end = '\r')

    print()
    return available, not_available




from setuptools import setup, find_packages
from venvgen.venvgen_version import __version__

with open('README.md', 'r') as f:
    description = f.read()


setup(
    name = 'venvgen',
    version = __version__,
    packages = find_packages(),
    install_requires = [
        'pandas',
        'inquirer',
        'tabulate',
        'requests'
    ],
    entry_points = {
        'console_scripts': [
            'venvgen = venvgen:main',
        ]
    },
    # include_package_data = True,
    # package_data = {
    #     'venvgen': ['database/*'],
    # },
    long_description = description,
    long_description_content_type = 'text/markdown',
)


from . import ANSI_color
from . import ANSI_line_modification
from . import sha256_generator
from . import system_control_protocol
from . import string_processing
from . import ui_helper
# from .ui_helper import introduction_screen, print_current_page



from . import OS
from .package_directory_manager import get_requirement_dir, create_requirement_dir
from .database_manager import select_the_latest_inserted_venv_id, get_specific_venv_connection, select_all_venv_list, update_requirement

# from ..general.ANSI_color import get_color_str

import re




# TODO: check why this thing yield error circular import
# from ..general.string_processing import remove_version_from_library


import importlib

import os

def update_install_requirement(venv_id):
    _, project_dir, venv_name, requirement_file = get_specific_venv_connection(venv_id)
    pip_freeze = OS.get_requirement(project_dir, venv_name)

    if requirement_file is None:
        requirement_dir = get_requirement_dir()
        requirement_file = os.path.join(requirement_dir, f'{venv_id}.txt')

    with open(requirement_file, 'w') as f:
        f.write(pip_freeze)

    return requirement_file

def check_requirement():
    create_requirement_dir()

    ## Testing
    venv_list = select_all_venv_list()
    requirement_dir = get_requirement_dir()

    for i in range(len(venv_list)):
        venv_id, venv_name, _, _, project_dir, _ = venv_list[i]
        is_connected, *_ = get_specific_venv_connection(venv_id)
        if is_connected:
            requirement_path = os.path.join(requirement_dir, f'{venv_id}.txt')
            update_install_requirement(venv_id)
            update_requirement(venv_id, requirement_path)

        percentage = ((i + 1) / len(venv_list))
        percentage_100 = percentage * 100

        MAX_BAR_LEN = 50

        percentage_bar_len = int(percentage * MAX_BAR_LEN) - 1

        remaining_bar_len = MAX_BAR_LEN - percentage_bar_len - 1

        percentage_bar = '=' * percentage_bar_len + '>'
        remaining_bar  = '.' * remaining_bar_len

        print(f'Progress: |{percentage_bar}{remaining_bar}| {percentage_100:.2f}%', end = '\r')



    ###

def requirement_generator(project_dir, venv):
    requirement_dir = get_requirement_dir()
    latest_inserted_id = select_the_latest_inserted_venv_id()
    pip_freeze = OS.get_requirement(project_dir, venv)

    with open(os.path.join(requirement_dir, f'{latest_inserted_id}.txt'), 'w') as f:
        f.write(pip_freeze)

    return latest_inserted_id, os.path.join(requirement_dir, f'{latest_inserted_id}.txt')

def requirement_is_updated(venv_id: int, project_dir: str, venv: str) -> bool:
    requirement_dir = get_requirement_dir()
    latest_inserted_id = select_the_latest_inserted_venv_id()
    pip_freeze = OS.get_requirement(project_dir, venv)

    with open(os.path.join(requirement_dir, f'{latest_inserted_id}.txt'), 'r') as f:
        requirement_file = f.read(pip_freeze)

    if pip_freeze == requirement_file:
        return True
    else:
        return False
    
def remove_version_from_library(library):
    pattern = r'==.*\n'
    pattern = re.compile(pattern, re.X)
    no_version_library = re.sub(pattern, '', library)
    return no_version_library


def get_libraries_from_venv_id(venv_id: int) -> list:
    requirement_file = os.path.join(get_requirement_dir(), f'{venv_id}.txt')

    libraries = []

    with open(requirement_file, 'r') as f:
        for line in f.readlines():
            line = remove_version_from_library(line)
            libraries.append(line)

    return libraries



def check_installed_libraries(venv_id: int, libraries: list) -> tuple[list, list]:

    not_installed = []
    installed = []

    requirement_libraries = get_libraries_from_venv_id(venv_id)

    for i in range(len(libraries)):
        
        if libraries[i] in requirement_libraries:
            installed.append(libraries[i])
        else:
            not_installed.append(libraries[i])

        percentage = ((i + 1) / len(libraries))
        percentage_100 = percentage * 100

        MAX_BAR_LEN = 50

        percentage_bar_len = int(percentage * MAX_BAR_LEN) - 1

        remaining_bar_len = MAX_BAR_LEN - percentage_bar_len - 1

        percentage_bar = '=' * percentage_bar_len + '>'
        remaining_bar  = '.' * remaining_bar_len

        print(f'Progress: |{percentage_bar}{remaining_bar}| {percentage_100:.2f}%', end = '\r')

    return installed, not_installed

